import tifffile
import os

from cellstitch_cuda.alignment import _label_overlap_cupy
from instanseg import InstanSeg
from cellstitch_cuda.postprocessing_cupy import (
    fill_holes_and_remove_small_masks,
    filter_nuclei_cells,
)

from cellstitch_cuda.alignment import *
from cellstitch_cuda.preprocessing_cupy import *


def relabel_layer(masks, z, lbls):
    """
    Relabel the label in LBLS in layer Z of MASKS.
    """
    layer = masks[z]
    if z != 0:
        reference_layer = masks[z - 1]
    else:
        reference_layer = masks[z + 1]

    overlap = _label_overlap_cupy(reference_layer, layer)

    for lbl in lbls:
        lbl0 = cp.argmax(overlap[:, lbl])
        layer[layer == lbl] = lbl0


def overseg_correction(masks):
    masks_cp = cp.asarray(masks)
    lbls = cp.unique(masks_cp)[1:]

    # get a list of labels that need to be corrected
    layers_lbls = {}

    for lbl in lbls:
        existing_layers = cp.any(masks_cp == lbl, axis=(1, 2))
        depth = cp.sum(existing_layers)

        if depth == 1:
            z = int(cp.where(existing_layers)[0][0])
            layers_lbls.setdefault(z, []).append(lbl)

    for z, lbls in layers_lbls.items():
        relabel_layer(masks_cp, z, lbls)
        cp._default_memory_pool.free_all_blocks()

    masks = masks_cp.get()
    cp._default_memory_pool.free_all_blocks()

    return masks


def full_stitch(
    xy_masks_prior,
    yz_masks,
    xz_masks,
    nuclei=None,
    filter: bool = True,
    n_jobs=-1,
    verbose=False,
):
    """Stitch masks in-place

    Stitches masks from top to bottom.

    Args:
        xy_masks_prior: numpy.ndarray with XY masks
        yz_masks: numpy.ndarray with YZ masks
        xz_masks: numpy.ndarray with XZ masks
        nuclei: numpy.ndarray with XY masks of nuclei
        filter: Use CellPose-based fill_holes_and_remove_small_masks() function. Default True
        n_jobs: Number of threads used. Set n_jobs to 1 for debugging parallel processing tasks. Default -1
        verbose: Verbosity. Default False
    """

    xy_masks = np.array(xy_masks_prior, dtype="uint32")
    num_frame = xy_masks.shape[0]
    prev_index = 0

    while Frame(xy_masks[prev_index]).is_empty():
        prev_index += 1

    curr_index = prev_index + 1

    time_start = time.time()

    while curr_index < num_frame:
        cp._default_memory_pool.free_all_blocks()
        if Frame(xy_masks[curr_index]).is_empty():
            # if frame is empty, skip
            curr_index += 1
        else:
            if verbose:
                print(
                    "===Stitching frame %s with frame %s ...==="
                    % (curr_index, prev_index)
                )

            yz_not_stitched = cp.asarray(
                (yz_masks[prev_index] != 0)
                * (yz_masks[curr_index] != 0)
                * (yz_masks[prev_index] != yz_masks[curr_index])
            )
            xz_not_stitched = cp.asarray(
                (xz_masks[prev_index] != 0)
                * (xz_masks[curr_index] != 0)
                * (xz_masks[prev_index] != xz_masks[curr_index])
            )

            fp = FramePair(
                xy_masks[prev_index], xy_masks[curr_index], max_lbl=xy_masks.max()
            )
            fp.stitch(yz_not_stitched, xz_not_stitched, verbose=verbose)
            xy_masks[curr_index] = fp.frame1.mask.get()

            prev_index = curr_index
            curr_index += 1

    if verbose:
        print("Total time to stitch: ", time.time() - time_start)

    if filter:
        cp._default_memory_pool.free_all_blocks()
        time_start = time.time()
        xy_masks = fill_holes_and_remove_small_masks(xy_masks, n_jobs=n_jobs)
        if verbose:
            print(
                "Time to fill holes and remove small masks: ", time.time() - time_start
            )

    cp._default_memory_pool.free_all_blocks()

    if not nuclei is None:
        time_start = time.time()
        xy_masks = filter_nuclei_cells(xy_masks, nuclei)
        cp._default_memory_pool.free_all_blocks()
        if verbose:
            print("Time to filter cells with nuclei: ", time.time() - time_start)

    time_start = time.time()

    xy_masks = overseg_correction(xy_masks)

    if verbose:
        print("Time to correct oversegmentation: ", time.time() - time_start)

    return xy_masks


def cellstitch_cuda(
    img,
    output_masks: bool = False,
    output_path=None,
    seg_mode: str = "nuclei_cells",
    pixel_size=None,
    z_step=None,
    bleach_correct: bool = True,
    filtering: bool = True,
    n_jobs: int = -1,
    verbose: bool = False,
):
    """All-in-one function to segment and stitch 2D labels

    Full stitching pipeline, which does the following:
        1. Histogram-based signal degradation correction
        2. Segmentation over the Z axis using InstanSeg
        3. Stitching of 2D planes into 3D labels through CellStitch's orthogonal labeling, which leverages Optimal
            Transport to create robust masks.

    Args:
        img: Either a path pointing to an existing image, or a numpy.ndarray. Must be 4D (ZCYX).
        output_masks: True to write all masks to the output path, or False to only return the final stitched mask.
            Default False
        output_path: Set to None to write to the input file location (if provided). Ignored of output_masks is False.
            Default None
        seg_mode: Instanseg segmentation mode: "nuclei" to only return nuclear masks, "cells" to return all the cell
            masks (including those without nuclei), or "nuclei_cells", which returns only cells with detected nuclei.
            Default "nuclei_cells"
        pixel_size: XY pixel size in microns per pixel. When set to None, will be read from img metadata if possible.
            Default None
        z_step: Z pixel size (z step) in microns per step. When set to None, will be read from img metadata if possible.
            Default None
        bleach_correct: Whether histogram-based signal degradation correction should be applied to img.
            Default True
        filtering: Whether the fill_holes_and_remove_small_masks function should be executed. With larger datasets, this
            has the tendency to massively slow down the postprocessing.
            Default True
        n_jobs: Set the number of threads to be used in parallel processing tasks. Use 1 for debugging. Generally, best
            left at the default value.
            Default -1
        verbose: Verbosity.
            Default False
    """

    # Check cuda
    if cp.cuda.is_available():
        print("CUDA is available. Using device", cp.cuda.get_device_id())
    else:
        print("CUDA is not available; using CPU.")

    # Initialize path
    path = ""

    # Read image file
    if os.path.isfile(img):
        path = str(img)
        with tifffile.TiffFile(path) as tif:
            img = tif.asarray()  # ZCYX
            tags = tif.pages[0].tags
    elif not isinstance(img, np.ndarray):
        print("img must either be a path to an existing image, or a numpy ndarray.")
        sys.exit(1)

    # Check image dimensions
    if img.ndim != 4:
        print("Expected a 4D image (ZCYX), while the img dimensions are ", img.ndim)
        sys.exit(1)

    # Set pixel sizes
    if pixel_size is None:
        try:
            pixel_size = 1 / (
                tags["XResolution"].value[0] / tags["XResolution"].value[1]
            )
            if verbose:
                print("Pixel size:", pixel_size)
        except:
            print(
                "No XResolution found in image metadata. The output might not be fully reliable."
            )
    if z_step is None:
        try:
            img_descr = tags["ImageDescription"].value.split()
            z_step = float(
                [s for s in img_descr if "spacing" in s][0].split("=")[-1]
            )  # At least it's pretty fast
            if verbose:
                print("Z step:", z_step)
        except:
            try:
                img_descr = tags["IJMetadata"].value["Info"].split()
                z_step = float(
                    [s for s in img_descr if "spacing" in s][0].split("=")[-1]
                )  # It's even funnier the second time
                if verbose:
                    print("Z step:", z_step)
            except:
                print(
                    "No spacing (Z step) found in image metadata. The output might not be fully reliable."
                )

    # Set up output path
    if output_masks:
        if output_path is None and os.path.isfile(path):
            output_path = os.path.split(path)[0]
        elif not os.path.exists(output_path):
            os.makedirs(output_path)

    # Instanseg-based pipeline
    model = InstanSeg("fluorescence_nuclei_and_cells")

    # Correct bleaching over Z-axis
    if bleach_correct:
        img = histogram_correct(img)
        cp._default_memory_pool.free_all_blocks()
        if verbose:
            print("Finished bleach correction.")

    img = img.transpose(1, 2, 3, 0)  # ZCYX -> CYXZ

    # Segment over Z-axis
    if verbose:
        print("Segmenting YX planes (Z-axis).")

    yx_masks = segmentation(img, model, seg_mode, xy=True)
    if seg_mode == "nuclei_cells":
        nuclei = yx_masks[1].transpose(2, 0, 1)  # YXZ -> ZYX
        yx_masks = yx_masks[0].transpose(2, 0, 1)  # YXZ -> ZYX

        if output_masks:
            tifffile.imwrite(os.path.join(output_path, "nuclei_masks.tif"), nuclei)
    else:
        yx_masks = yx_masks.transpose(2, 0, 1)  # YXZ -> ZYX

    if torch.cuda.is_available():
        torch.cuda.empty_cache()  # Clear GPU cache

    if output_masks:
        tifffile.imwrite(os.path.join(output_path, "yx_masks.tif"), yx_masks)

    # Segment over X-axis
    if verbose:
        print("Segmenting YZ planes (X-axis).")
    transposed_img = img.transpose(0, 1, 3, 2)  # CYXZ -> CYZX
    transposed_img = upscale_img(
        transposed_img, pixel_size, z_step
    )  # Preprocess YZ planes
    cp._default_memory_pool.free_all_blocks()
    yz_masks = segmentation(transposed_img, model, seg_mode)
    del transposed_img
    if torch.cuda.is_available():
        torch.cuda.empty_cache()  # Clear GPU cache
    yz_masks = downscale_mask(yz_masks, pixel_size, z_step).transpose(
        1, 0, 2
    )  # YZX -> ZYX
    cp._default_memory_pool.free_all_blocks()
    if output_masks:
        tifffile.imwrite(os.path.join(output_path, "yz_masks.tif"), yz_masks)

    # Segment over Y-axis
    if verbose:
        print("Segmenting XZ planes (Y-axis).")
    transposed_img = img.transpose(0, 2, 3, 1)  # CYXZ -> CXZY
    transposed_img = upscale_img(
        transposed_img, pixel_size, z_step
    )  # Preprocess XZ planes
    cp._default_memory_pool.free_all_blocks()
    xz_masks = segmentation(transposed_img, model, seg_mode)
    if torch.cuda.is_available():
        torch.cuda.empty_cache()  # Clear GPU cache
    xz_masks = downscale_mask(xz_masks, pixel_size, z_step).transpose(
        1, 2, 0
    )  # XZY -> ZYX
    cp._default_memory_pool.free_all_blocks()
    if output_masks:
        tifffile.imwrite(os.path.join(output_path, "xz_masks.tif"), xz_masks)

    # Memory cleanup
    del model, img, transposed_img
    if torch.cuda.is_available():
        torch.cuda.empty_cache()  # Clear GPU cache

    if verbose:
        print("Running CellStitch stitching...")

    if seg_mode == "nuclei_cells":
        cellstitch_masks = full_stitch(
            yx_masks,
            yz_masks,
            xz_masks,
            nuclei,
            filter=filtering,
            n_jobs=n_jobs,
            verbose=verbose,
        )
    else:
        cellstitch_masks = full_stitch(
            yx_masks,
            yz_masks,
            xz_masks,
            filter=filtering,
            n_jobs=n_jobs,
            verbose=verbose,
        )

    cp._default_memory_pool.free_all_blocks()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()  # Clear GPU cache

    if output_masks:
        tifffile.imwrite(
            os.path.join(output_path, "cellstitch_masks.tif"), cellstitch_masks
        )

    return cellstitch_masks



import tifffile
import os
from cellpose.utils import stitch3D
from cellstitch_cuda.pipeline import full_stitch


stitch_method = "cellstitch"  # "iou" or "cellstitch"
file_path_yx_masks = r"E:\1_DATA\Rheenen\tvl_jr\3d stitching\raw4-large\yx_masks.tif"
file_path_yz_masks = r"E:\1_DATA\Rheenen\tvl_jr\3d stitching\raw4-large\yz_masks.tif"
file_path_xz_masks = r"E:\1_DATA\Rheenen\tvl_jr\3d stitching\raw4-large\xz_masks.tif"
file_path_nuclei_masks = r"E:\1_DATA\Rheenen\tvl_jr\3d stitching\raw4-large\nuclei_masks.tif"
out_path = os.path.split(file_path_yx_masks)[0]

# Read YX masks
yx_masks = tifffile.imread(file_path_yx_masks)

if stitch_method == "iou":

    print("Running IoU stitching...")

    iou_masks = stitch3D(yx_masks, stitch_threshold=0.25)
    tifffile.imwrite(os.path.join(out_path, "iou_masks.tif"), iou_masks)

elif stitch_method == "cellstitch":

    # Read YZ masks
    yz_masks = tifffile.imread(file_path_yz_masks)

    # Read XZ masks
    xz_masks = tifffile.imread(file_path_xz_masks)

    nuclei = tifffile.imread(file_path_nuclei_masks)

    print("Running CellStitch stitching...")

    cellstitch_masks = full_stitch(
        yx_masks, yz_masks, xz_masks, nuclei, verbose=True
    )

    tifffile.imwrite(os.path.join(out_path, "cellstitch_masks.tif"), cellstitch_masks)

else:
    print("Incompatible stitching method.")



import ot
import ot.plot
import cupy as cp
from cupyx.scipy import ndimage as ndi

from cellpose import utils as cp_utils


# -------------------------------
# interpolation helper functions
# -------------------------------


def comp_match_plan(pc1, pc2, dist="sqeuclidean"):
    """Compute optimal matching plans between 2 sets of point clouds"""
    # compute cost matrix
    C = ot.dist(pc1, pc2, metric=dist).astype(cp.float64)
    C /= C.max()

    # convert point clouds to uniform distributions
    n_pts1, n_pts2 = pc1.shape[0], pc2.shape[0]
    mu1, mu2 = cp.ones(n_pts1) / n_pts1, cp.ones(n_pts2) / n_pts2

    # compute transport plan
    plan = ot.emd(mu1, mu2, C)

    return plan


def interpolate(pc1, pc2, dist="sqeuclidean", anisotropy=2):
    """
    Calculate interpolated predictions

    Parameters
    ----------
    pc1 : cp.ndarray
        Point cloud representing cell boundary in frame 1

    pc2 : cp.ndarray
        Point cloud representing cell boundary in frame 2

    anisotropy : int
        Ratio of sampling rate between different axes

    Returns
    -------
    interp_pcs : list
        Smoothed boundary locations along interpolated layers
    """
    alphas = cp.linspace(0, 1, anisotropy + 1)[1:-1]
    plan = comp_match_plan(pc1, pc2, dist=dist)
    normalized_plan = plan / plan.sum(
        axis=1, keepdims=1
    )  # normalize so that the row sum is 1

    interp_pcs = []

    for alpha in alphas:
        n_pts = pc1.shape[0]
        avg_pc = cp.zeros((n_pts, 2), dtype=int)

        for i in range(n_pts):
            point = pc1[i]
            target_weights = normalized_plan[i]

            weighted_target = cp.array(
                [cp.sum(target_weights * pc2[:, 0]), cp.sum(target_weights * pc2[:, 1])]
            )

            avg_pc[i, :] = point * (1 - alpha) + alpha * weighted_target

        interp_pcs.append(avg_pc)

    return interp_pcs


# -----------------------------------------
# Util functions for interp reconstruction
# -----------------------------------------


def get_lbls(mask):
    """Get unique labels from the predicted masks"""
    return cp.unique(mask)[1:]


def min_size_filter(res, thld=100):
    """Filter out all masks with area below threshold"""
    assert len(res) == 4, "Invalid Cellpose 4-tuple result"
    preds = res[0]

    for i in range(len(preds)):
        lbls = get_lbls(preds[i])
        for lbl in lbls:
            msk = preds[i] == lbl
            if msk.sum() < thld:
                coords = cp.nonzero(msk)
                preds[i][coords] = 0

    res_filtered = (preds, res[1], res[2], res[3])
    return res_filtered


def get_contours(masks):
    """Transfer solid mask predictions to non-overlapping contours w/ distinct integers"""
    masks_new = masks.copy()
    outlines = cp_utils.masks_to_outlines(masks)
    masks_new[~outlines] = 0

    return masks_new


def get_mask_perimeter(masks, lbl, is_contour=False):
    assert lbl in get_lbls(
        masks
    ), "Label {} doesn't in current mask predictions".format(lbl)

    if is_contour:
        p = (masks == lbl).sum()
    else:
        mask_lbl = (masks == lbl).astype(cp.uint8)
        p = cp_utils.masks_to_outlines(mask_lbl).sum()

    return p


def calc_vols(pred):
    """
    Calculate volumes of each mask from predictions
    """
    lbls = get_lbls(pred)
    vols = [(pred == lbl).sum() for lbl in lbls]
    return vols


def calc_depth(masks):
    """
    Calculate z-layer depth of predictions
    """
    assert masks.ndim == 3, "Mask predictions must be 3D to calculate depth"

    lbls = get_lbls(masks)
    depths = cp.vectorize(lambda lbl: cp.diff(cp.nonzero(masks == lbl)[0][[0, -1]])[0])(
        lbls
    )

    return depths


def mask_to_coord(mask):
    """Return (n, 2) coordinates from masks"""
    coord = cp.asarray(cp.nonzero(mask)).T
    return coord


def coord_to_mask(coord, size, lbl):
    """Convert from coordinates to original labeled masks"""
    mask = cp.zeros(size)
    mask[tuple(coord.T)] = lbl
    return mask


def contour_to_mask(contour):
    lbl = get_lbls(contour)[0]
    """ Convert contour to solid masks with fill-in labels"""
    binary_contour = contour > 0
    binary_mask = ndi.binary_fill_holes(binary_contour)

    mask = cp.zeros_like(binary_contour)
    mask[binary_mask] = lbl

    return mask


def connect(coord1, coord2, mask):
    """
    Modify the mask by connecting the two given coordinates.

    Parameters
    ----------
    coord1 : [x1, y1]
        Coordinate of the first pixel.

    coord2 : [x2, y2]
        Coordinate of the second pixel.

    mask : binary cp.narray
        Binary mask of the boundary.
    """
    x1, y1 = coord1
    x2, y2 = coord2

    x_offset, y_offset = x2 - x1, y2 - y1

    # skip if the two coordinates are already connected
    if x_offset**2 + y_offset**2 <= 2:
        return

    diag_length = min(abs(x_offset), abs(y_offset))

    # initialize at coord1
    added_x, added_y = x1, y1

    # first, add diagonal pixels
    for i in range(1, diag_length + 1):
        added_x = x1 + i * cp.sign(x_offset)
        added_y = y1 + i * cp.sign(y_offset)

        mask[added_x, added_y] = 1

        # need to walk vertically
    if added_x == x2 and added_y != y2:
        offset = abs(added_y - y2)
        for i in range(1, offset + 1):
            mask[added_x, added_y + i * cp.sign(y_offset)] = 1

            # or, now need to walk horizonally
    if added_y == y2 and added_x != x2:
        offset = abs(added_x - x2)
        for i in range(1, offset + 1):
            mask[added_x + i * cp.sign(x_offset), added_y] = 1


def calc_angles(sc_pt, tg_pts, eps=1e-20):
    """
    Calculate angle (rad) between source point (sc_pt) & list of target points(tg_pts, dim=(n, 2))
    """
    sc_pts = cp.tile(sc_pt, (tg_pts.shape[0], 1))
    diff = tg_pts - sc_pts
    angles = cp.apply_along_axis(
        lambda x: cp.arctan2(x[1], x[0] + eps), axis=1, arr=diff
    )

    return angles


def connect_boundary(coords, size, lbl=1):
    """
    Connect interpolation coordinates to generate close-loop mask contours

    Parameters
    ----------
    coords : ndarray, shape (ns,2)
        Boundary coordinates (might be disconnected).

    size: (n1, n2)
        Shape of the final mask.

    lbl: int
        Label of the original mask.

    Returns
    -------
    mask: ndarray, shape (n1, n2)
        Connected boundary mask.

    """
    # Sort boundary labels by angle to mask's mass center
    mass_center = cp.round(coords.mean(0)).astype(cp.int64)
    angles = calc_angles(mass_center, coords)
    sorted_coords = coords[angles.argsort()]

    mask = coord_to_mask(coords, size, lbl)

    for i, (x, y) in enumerate(sorted_coords[:-1]):
        next_x, next_y = sorted_coords[i + 1]
        connect((x, y), (next_x, next_y), mask)

    connect(tuple(sorted_coords[-1]), tuple(sorted_coords[0]), mask)

    return mask


# -----------------------------
# Core Interpolation functions
# -----------------------------


def interp_layers(sc_mask, tg_mask, dist="sqeuclidean", anisotropy=2):
    """
    Interpolating adjacent z-layers
    """

    def _dilation(coords, lims):
        y, x = coords
        ymax, xmax = lims
        dy, dx = cp.meshgrid(
            cp.arange(y - 2, y + 3), cp.arange(x - 2, x + 3), indexing="ij"
        )
        dy, dx = dy.flatten(), dx.flatten()
        mask = cp.logical_and(
            cp.logical_and(dy >= 0, dx >= 0), cp.logical_and(dy < ymax, dx < xmax)
        )
        return dy[mask], dx[mask]

    shape = sc_mask.shape
    sc_contour = get_contours(sc_mask)
    tg_contour = get_contours(tg_mask)

    # Boundary condition: if empty on source / target label
    # align the empty slice w/ mass centers to represent instance endings
    sc_dummy = cp.zeros_like(sc_mask)
    tg_dummy = cp.zeros_like(tg_mask)
    if not cp.intersect1d(get_lbls(sc_contour), get_lbls(tg_contour)).size:
        if (sc_contour.sum() == tg_contour.sum() == 0) or (
            cp.logical_and(sc_mask, tg_mask).sum() > 0
        ):
            return cp.zeros(shape)
        get_mask_center = lambda x: (
            cp.round(cp.nonzero(x)[0].sum() / x.sum()).astype(cp.uint16),
            cp.round(cp.nonzero(x)[1].sum() / x.sum()).astype(cp.uint16),
        )
        for lbl in get_lbls(sc_contour):
            yc, xc = _dilation(get_mask_center(sc_mask == lbl), sc_mask.shape)
            tg_dummy[yc, xc] = lbl
        for lbl in get_lbls(tg_contour):
            yc, xc = _dilation(get_mask_center(tg_mask == lbl), tg_mask.shape)
            sc_dummy[yc, xc] = lbl
        sc_contour += sc_dummy
        tg_contour += tg_dummy

    joint_lbls = cp.intersect1d(get_lbls(sc_contour), get_lbls(tg_contour))

    interp_masks = cp.zeros(
        (
            anisotropy + 1,  # num. interpolated layers
            len(joint_lbls),  # num. individual masks
            shape[0],  # x
            shape[1],  # y
        )
    )

    for i, lbl in enumerate(joint_lbls):
        sc_ct = (sc_contour == lbl).astype(cp.uint8)
        tg_ct = (tg_contour == lbl).astype(cp.uint8)

        sc_coord = mask_to_coord(sc_ct)
        tg_coord = mask_to_coord(tg_ct)

        interp_coords = interpolate(
            sc_coord, tg_coord, dist=dist, anisotropy=anisotropy
        )
        interps = [
            ndi.binary_fill_holes(connect_boundary(interp, shape)) * lbl
            for interp in interp_coords
        ]

        interp_masks[1:-1, i, ...] = interps

    interp_masks = interp_masks.max(1)
    interp_masks[0] = sc_mask
    interp_masks[-1] = tg_mask

    return interp_masks


def full_interpolate(masks, anisotropy=2, dist="sqeuclidean", verbose=False):
    """
    Interpolating between all adjacent z-layers

    Parameters
    ----------
    masks : cp.ndarray
        layers of 2D predictions
        (dim: (Depth, H, W))

    anisotropy : int
        Ratio of sampling rate between xy-axes & z-axis

    Returns
    -------|
    interp_masks : cp.ndarray
        interpolated masks
        (dim: (Depth * anisotropy - (anisotropy-1), H, W))

    """
    interp_masks = cp.zeros(
        (
            len(masks) + (len(masks) - 1) * (anisotropy - 1),
            masks.shape[1],
            masks.shape[2],
        )
    )

    idx = 0
    for i, sc_mask in enumerate(masks[:-1]):
        if verbose and i % 20 == 0:
            print("Interpolating layer {} & {}...".format(i, i + 1))
        tg_mask = masks[i + 1]
        interps = interp_layers(sc_mask, tg_mask, dist=dist, anisotropy=anisotropy)
        interp_masks[idx : idx + anisotropy + 1] = interps
        idx += anisotropy

    return interp_masks



import ot
import cupyx
from skimage import color
import matplotlib.pyplot as plt

from cellstitch_cuda.frame import *
import time


class FramePair:
    def __init__(self, mask0, mask1, max_lbl=0):
        self.frame0 = Frame(mask0)
        self.frame1 = Frame(mask1)

        # store the max labels for stitching
        max_lbl_default = max(
            self.frame0.get_lbls().max(), self.frame1.get_lbls().max()
        )

        self.max_lbl = max(max_lbl, max_lbl_default)

    def display(self):
        """
        Display frame0 and frame1 next to each other, with consistent colorings.
        """

        num_lbls = len(cp.union1d(self.frame0.get_lbls(), self.frame1.get_lbls()))

        colors = cp.random.random((num_lbls, 3))

        frames = cp.array([self.frame0.mask, self.frame1.mask])
        rgb = color.label2rgb(frames, colors=colors, bg_label=0)

        fig, axes = plt.subplots(1, 2, figsize=(20, 10))
        axes[0].imshow(rgb[0])
        axes[1].imshow(rgb[1])

        plt.tight_layout()
        plt.show()

    def get_plan(self, C):
        """
        Compute the transport plan between the two frames, given the cost matrix between the cells.
        """
        mask0 = cp.asarray(self.frame0.mask)
        mask1 = cp.asarray(self.frame1.mask)

        unique_labels0, counts0 = cp.unique(mask0, return_counts=True)
        unique_labels1, counts1 = cp.unique(mask1, return_counts=True)
        sizes0 = cp.asarray(counts0)
        sizes1 = cp.asarray(counts1)

        # convert to distribution to compute transport plan
        dist0 = sizes0 / cp.sum(sizes0)
        dist1 = sizes1 / cp.sum(sizes1)

        # compute transportation plan
        plan = ot.emd(dist0, dist1, C)

        return plan, mask1

    def get_cost_matrix(self, overlap, lbls0, lbls1):
        """
        Return the cost matrix between cells in the two frame defined by IoU.
        """

        sizes0 = cp.sum(overlap, axis=1)
        sizes1 = cp.sum(overlap, axis=0)

        # Create a meshgrid for vectorized operations
        lbl0_indices, lbl1_indices = cp.meshgrid(lbls0, lbls1, indexing="ij")

        overlap_sizes = overlap[lbl0_indices, lbl1_indices]
        scaling_factors = overlap_sizes / (
            sizes0[lbl0_indices] + sizes1[lbl1_indices] - overlap_sizes
        )

        C = 1 - scaling_factors

        return C

    def stitch(
        self, yz_not_stitched, xz_not_stitched, p_stitching_votes=0.75, verbose=False
    ):
        """Stitch frame1 using frame 0."""

        time_start = time.time()

        lbls0 = self.frame0.get_lbls()  # Get unique label IDs
        lbls1 = self.frame1.get_lbls()  # Get unique label IDs

        # get sizes
        overlap = _label_overlap_cupy(self.frame0.mask, self.frame1.mask)

        # compute matching
        C = self.get_cost_matrix(overlap, lbls0, lbls1)
        plan, mask1 = self.get_plan(C)

        # get a soft matching from plan
        n, m = plan.shape
        soft_matching = cp.zeros((n, m))

        # Vectorized computation
        matched_indices = plan.argmax(axis=1)
        soft_matching[cp.arange(n), matched_indices] = 1

        stitched_mask1 = cp.zeros(mask1.shape)
        for lbl1_index in range(1, m):
            # find the cell with the lowest cost (i.e. lowest scaled distance)
            matching_filter = soft_matching[:, lbl1_index]
            filtered_C = cp.where(
                matching_filter == 0, cp.Inf, C[:, lbl1_index]
            )  # ignore the non-matched cells

            lbl0_index = cp.argmin(
                filtered_C
            )  # this is the cell0 we will attempt to relabel cell1 with

            lbl0, lbl1 = int(lbls0[lbl0_index]), int(lbls1[lbl1_index])

            n_not_stitch_pixel = (
                yz_not_stitched[mask1 == lbl1].sum() / 2
                + xz_not_stitched[mask1 == lbl1].sum() / 2
            )
            stitch_cell = (
                n_not_stitch_pixel <= (1 - p_stitching_votes) * (mask1 == lbl1).sum()
            )

            if lbl0 != 0 and stitch_cell:  # only reassign if they overlap
                stitched_mask1[mask1 == lbl1] = lbl0
            else:
                self.max_lbl += 1
                stitched_mask1[mask1 == lbl1] = self.max_lbl  # create a new label

        if verbose:
            print("Time to stitch: ", time.time() - time_start)

        self.frame1 = Frame(stitched_mask1)


def _label_overlap_cupy(x, y):
    """Fast function to get pixel overlaps between masks in x and y.

    Args:
        x (np.ndarray, int): Where 0=NO masks; 1,2... are mask labels.
        y (np.ndarray, int): Where 0=NO masks; 1,2... are mask labels.

    Returns:
        overlap (np.ndarray, int): Matrix of pixel overlaps of size [x.max()+1, y.max()+1].

    Adapted from CellPose: https://github.com/MouseLand/cellpose
        https://doi.org/10.1038/s41592-020-01018-x: Stringer, C., Wang, T., Michaelos, M., & Pachitariu, M. (2021).
        Cellpose: a generalist algorithm for cellular segmentation. Nature methods, 18(1), 100-106.
        Copyright © 2023 Howard Hughes Medical Institute, Authored by Carsen Stringer and Marius Pachitariu.
    """
    # put label arrays into standard form then flatten them
    x = cp.asarray(x.ravel())
    y = cp.asarray(y.ravel())
    xmax = int(x.max())
    ymax = int(y.max())

    # preallocate a "contact map" matrix
    overlap = cp.zeros((1 + xmax, 1 + ymax), dtype=cp.uint)

    cupyx.scatter_add(overlap, (x, y), 1)

    return overlap



# import tifffile
from cellstitch_cuda.pipeline import cellstitch_cuda
# from skimage.data import cells3d
#
# im = cells3d()
# 0.29 0.26 0.26
img = r"E:\1_DATA\Rheenen\tvl_jr\3d stitching\unmixed2\unmixed.tif"

masks = cellstitch_cuda(img, output_masks=True, verbose=True, seg_mode="nuclei_cells", n_jobs=-1)
